import sys, os, argparse, pickle, copy, random, string, csv, warnings
import numpy as np
import pandas as pd
import scanpy as sc
from tqdm import tqdm
from pathlib import Path
from scipy import linalg as LA
from scipy.linalg import LinAlgWarning
from scipy.spatial.transform import Rotation as R
from cytocraft import model
from cytocraft import util
from cytocraft.stereopy import *
from cytocraft.rigid import *
import warnings

warnings.filterwarnings("ignore")

"""
This module implements main functions.
"""


def genedistribution(gem, CellIDs, TopGenes):
    W = np.zeros((len(CellIDs) * 2, len(TopGenes)))
    i = 0
    for c in CellIDs:
        j = 0
        # subset the gem in terms of individual cells
        gem_cell = gem[gem.CellID == c]
        for n in TopGenes:
            if n not in gem_cell.geneID.values:
                x_median = np.nan
                y_median = np.nan
            else:
                gem_cell_gene = gem_cell[gem_cell.geneID == n]
                x_median = np.average(
                    gem_cell_gene.x.values.astype(int),
                    weights=gem_cell_gene.MIDCount.values.astype(float),
                )
                y_median = np.average(
                    gem_cell_gene.y.values.astype(int),
                    weights=gem_cell_gene.MIDCount.values.astype(float),
                )
            W[i * 2, j] = x_median
            W[i * 2 + 1, j] = y_median
            j += 1
        i += 1
    return W


def MASK(gem, GeneIDs, CellIDs, Ngene):
    mask = np.zeros((len(CellIDs), len(GeneIDs)))
    i = 0
    for c in CellIDs:
        gem_cell = gem[gem.CellID == c]
        GeneUIDs_cell = (
            gem_cell.groupby(["geneID"])["MIDCount"]
            .count()
            .reset_index(name="Count")
            .sort_values(["Count"], ascending=False)
            .geneID[:Ngene]
        )

        mask[i] = np.isin(GeneIDs, GeneUIDs_cell)
        i += 1
    return np.bool_(mask)


def DeriveRotation(W, X, Mask, CellUIDs, adata):
    F = int(W.shape[0] / 2)
    Rotation = np.zeros((F, 3, 3))
    pop_indices = []
    for i in range(F):
        try:
            # print("derive rotation for cell " + str(i))
            # select genes
            Wi = W[:, Mask[i, :]]
            # filter cells
            Wi_filter = Wi[~np.isnan(Wi).any(axis=1), :]
            while Wi_filter.shape[0] < 6:
                # reduce the numebr of gene in Mask if cell number less than 3
                Mask[i, :] = change_last_true(Mask[i, :])
                # print("reduce one gene for cell " + str(i))
                # filter genes
                Wi = W[:, Mask[i, :]]
                # filter cells
                Wi_filter = Wi[~np.isnan(Wi).any(axis=1), :]
            idx = int(find_subarray(Wi_filter, Wi[i * 2]) / 2)
            Xi = X[Mask[i, :], :]
            model = factor(Wi_filter)
            _, R, _ = numpy_svd_rmsd_rot(np.dot(model.Rs[idx], model.Ss[0]).T, Xi)
            Rotation[i] = R
        except Exception as err:
            print("Cell No." + str(i + 1) + " is removed due to: ", err)
            pop_indices.append(i)
    for i in sorted(pop_indices, reverse=True):
        Rotation = np.delete(Rotation, obj=i, axis=0)
        W = np.delete(W, obj=i * 2 + 1, axis=0)
        W = np.delete(W, obj=i * 2, axis=0)
        adata = adata[~(adata.obs.index.values == str(CellUIDs[i])), :]
        del CellUIDs[i]
    return Rotation, W, CellUIDs, adata


def UpdateX(RM, W, GeneUID, *X_old):
    F = int(W.shape[0] / 2)
    pop_indices = []
    for j in range(W.shape[1]):
        a1 = b1 = c1 = d1 = a2 = b2 = c2 = d2 = a3 = b3 = c3 = d3 = 0
        for i in range(F):
            if not (
                np.isnan(W[i * 2 : i * 2 + 2, j]).any() or np.isnan(RM[i, :, :]).any()
            ):
                a1 += RM[i, 0, 0] * RM[i, 0, 0] + RM[i, 0, 1] * RM[i, 0, 1]
                b1 += RM[i, 0, 0] * RM[i, 1, 0] + RM[i, 0, 1] * RM[i, 1, 1]
                c1 += RM[i, 0, 0] * RM[i, 2, 0] + RM[i, 0, 1] * RM[i, 2, 1]

                a2 += RM[i, 1, 0] * RM[i, 0, 0] + RM[i, 1, 1] * RM[i, 0, 1]
                b2 += RM[i, 1, 0] * RM[i, 1, 0] + RM[i, 1, 1] * RM[i, 1, 1]
                c2 += RM[i, 1, 0] * RM[i, 2, 0] + RM[i, 1, 1] * RM[i, 2, 1]

                a3 += RM[i, 2, 0] * RM[i, 0, 0] + RM[i, 2, 1] * RM[i, 0, 1]
                b3 += RM[i, 2, 0] * RM[i, 1, 0] + RM[i, 2, 1] * RM[i, 1, 1]
                c3 += RM[i, 2, 0] * RM[i, 2, 0] + RM[i, 2, 1] * RM[i, 2, 1]

                d1 += RM[i, 0, 0] * W[i * 2, j] + RM[i, 0, 1] * W[i * 2 + 1, j]
                d2 += RM[i, 1, 0] * W[i * 2, j] + RM[i, 1, 1] * W[i * 2 + 1, j]
                d3 += RM[i, 2, 0] * W[i * 2, j] + RM[i, 2, 1] * W[i * 2 + 1, j]
            else:
                # print("skip cell" + str(i) + "for gene" + str(j))
                pass

        args = np.array([[a1, b1, c1], [a2, b2, c2], [a3, b3, c3]])
        results = np.array([d1, d2, d3])
        try:
            newXi = LA.solve(args, results)
            try:
                newX = np.append(
                    newX,
                    [newXi],
                    axis=0,
                )
            except NameError:
                newX = np.array([newXi])
        except (np.linalg.LinAlgError, LinAlgWarning) as err:
            print("Gene No." + str(j + 1) + " is removed due to: ", err)
            pop_indices.append(j)

    # Drop genes
    ## X_old also need to drop genes, so that Xs can be compared
    if X_old:
        X_old = X_old[0]
        for i in sorted(pop_indices, reverse=True):
            del GeneUID[i]
            W = np.delete(W, obj=i, axis=1)
            X_old = np.delete(X_old, obj=i, axis=0)
        return newX, W, GeneUID, X_old
    else:
        for i in sorted(pop_indices, reverse=True):
            del GeneUID[i]
            W = np.delete(W, obj=i, axis=1)
        return newX, W, GeneUID


def numpy_svd_rmsd_rot(in_crds1, in_crds2):
    """
    Returns rmsd and optional rotation between 2 sets of [nx3] arrays.

    This requires numpy for svd decomposition.
    The transform direction: transform(m, ref_crd) => target_crd.
    """

    crds1 = np.array(in_crds1)
    crds2 = np.array(in_crds2)
    assert crds1.shape[1] == 3
    assert crds1.shape == crds2.shape

    n_vec = np.shape(crds1)[0]
    correlation_matrix = np.dot(np.transpose(crds1), crds2)
    v, s, w = np.linalg.svd(correlation_matrix)
    is_reflection = (np.linalg.det(v) * np.linalg.det(w)) < 0.0

    if is_reflection:
        s[-1] = -s[-1]
    E0 = sum(sum(crds1 * crds1)) + sum(sum(crds2 * crds2))
    rmsd_sq = (E0 - 2.0 * sum(s)) / float(n_vec)
    rmsd_sq = max([rmsd_sq, 0.0])
    rmsd = np.sqrt(rmsd_sq)

    if is_reflection:
        v[-1, :] = -v[-1, :]
    rot33 = np.dot(v, w).transpose()
    # print(is_reflection)
    return rmsd, rot33, is_reflection


def normalizeW(W):
    result = np.empty_like(W)
    for i in range(int(W.shape[0])):
        result[i] = W[i] - np.nanmean(W[i])
    return result


def filterGenes(array, Genes, threshold):
    # threshold is a float between 0 and 1, indicating the maximum proportion of np.nans allowed in a column
    # returns a new array with only the columns that have less than threshold proportion of np.nans
    # if array is empty or threshold is invalid, returns None

    # check if array is empty
    if array.size == 0:
        return None

    # check if threshold is valid
    if not (0 <= threshold <= 1):
        return None

    # get the number of rows and columns in the array
    rows, cols = array.shape

    # create a list to store the indices of the columns to keep
    keep_cols = []

    # loop through each column
    for i in range(cols):
        # get the column as a 1D array
        col = array[:, i]

        # count the number of np.nans in the column
        nan_count = np.count_nonzero(np.isnan(col))

        # calculate the proportion of np.nans in the column
        nan_prop = nan_count / rows

        # if the proportion is less than the threshold, add the index to the list
        if nan_prop < threshold:
            keep_cols.append(i)

    # create a new array with only the columns in the list
    new_array = array[:, keep_cols]
    Genes = list(np.array(Genes)[keep_cols])

    # return the new array
    return new_array, Genes


def gene_chr(species):
    if species == "Mouse" or species == "Mice":
        gtf_file = (
            os.path.dirname(os.path.realpath(__file__)) + "/gtf/mice_genes.gene.gtf"
        )
    elif species == "Axolotls" or species == "Axolotl":
        gtf_file = (
            os.path.dirname(os.path.realpath(__file__))
            + "/gtf/AmexT_v47-AmexG_v6.0-DD.gene.gtf"
        )
    elif species == "Human" or species == "Monkey":
        gtf_file = (
            os.path.dirname(os.path.realpath(__file__))
            + "/gtf/gencode.v44.basic.annotation.gene.gtf"
        )
    gene_chr = {}
    with open(gtf_file, "r") as f:
        for line in f:
            if not line.startswith("#"):
                inf = line.strip().split("\t")
                if inf[2] == "gene":
                    chrom = inf[0]
                    if chrom not in gene_chr.keys():
                        gene_chr[chrom] = []
                    gene_names = inf[8].split(";")
                    try:
                        if (
                            species == "Mouse"
                            or species == "Mice"
                            or species == "Human"
                            or species == "Monkey"
                        ):
                            gene_symbol = gene_names[2].split()[1].strip('"')
                        elif species == "Axolotls":
                            gene_symbol = gene_names[1].split()[1].strip('"')
                        gene_chr[gene_symbol] = chrom
                    except IndexError:
                        print("IndexError")
                        pass
    return gene_chr


def gene_gene_distance_matrix(X):
    GeneList = X.index
    N = len(GeneList)
    DM = np.zeros((N, N))
    for n, _ in tqdm(enumerate(GeneList)):
        for m, _ in enumerate(GeneList[: n + 1]):
            d = np.linalg.norm(X.iloc[n] - X.iloc[m])
            DM[n, m] = DM[m, n] = d
    return DM


def RMSD_distance_matrix(Xs, GeneLists, keys, ngene=100):
    N = len(keys)
    DM = np.zeros((N, N))

    for n, key_n in tqdm(enumerate(keys)):
        for m, key_m in enumerate(keys[: n + 1]):
            intersected_values = np.intersect1d(GeneLists[key_n], GeneLists[key_m])[
                :ngene
            ]
            boolean_arrays_n = np.in1d(GeneLists[key_n], intersected_values)
            boolean_arrays_m = np.in1d(GeneLists[key_m], intersected_values)
            normalized_x_n = normalizeX(Xs[key_n][boolean_arrays_n], method="mean")
            normalized_x_m = normalizeX(Xs[key_m][boolean_arrays_m], method="mean")
            d1, _, _ = numpy_svd_rmsd_rot(normalized_x_n, normalized_x_m)
            d2, _, _ = numpy_svd_rmsd_rot(mirror(normalized_x_n), normalized_x_m)
            DM[n, m] = DM[m, n] = min(d1, d2)
    return DM


def mirror(X):
    mirrorX = np.copy(X)
    mirrorX[:, 2] = -mirrorX[:, 2]
    return mirrorX


def euclidean_distance(coord1, coord2):
    distance = math.sqrt(sum([(x1 - x2) ** 2 for x1, x2 in zip(coord1, coord2)]))
    return distance


def knn_neighbors(X, k):
    if k != 0:
        knn_m = np.zeros((X.shape[0], X.shape[0]))
        for i in range(X.shape[0]):
            ids = np.argpartition(X[i], -k)[-k:]
            top_set = set(X[i, ids])
            if len(top_set) == 1:
                b = X[i] == top_set.pop()
                ids = []
                offset = 1
                left = True
                while (len(ids) < k) and (offset < X.shape[0]):
                    # while len(ids) < k:
                    if left:
                        idx = i + offset
                    else:
                        idx = i - offset
                    if idx < 0 or idx > len(b) - 1:
                        offset += 1
                        left = not left
                        continue
                    if b[idx]:
                        ids.append(idx)
                    offset += 1
                    left = not left
            knn_m[i, ids] = 1

        knn_m = (knn_m + knn_m.T) / 2
        knn_m[np.nonzero(knn_m)] = 1
    else:
        knn_m = X
    return knn_m


def write_pdb(show_X, genechr, geneLst, write_path, sp, seed, prefix="chain"):
    if sp == "Axolotls":
        uniquechains = [
            chr for chr in sorted(set(genechr.keys())) if chr.startswith("chr")
        ]
    else:
        uniquechains = sorted(set(genechr.keys()))

    # i is global gene index in genome include terminals
    # j is global gene index in genome
    i = 1
    j = 1
    chain_idx = 0
    for chain in uniquechains:
        k = 0
        show_shape = show_X.T
        rows = []
        rows.append(
            "HEADER".ljust(10, " ")
            + "CHROMOSOMES".ljust(40, " ")
            + get_date_today()
            + "   "
            + str(seed).ljust(4, " ")
            + "\n"
        )
        rows.append(
            "TITLE".ljust(10, " ") + "CHROMOSOME CONFORMATION INFERED FROM sST DATA\n"
        )
        Is = []
        for g in genechr[chain]:
            if g in geneLst:
                idx = geneLst.index(g)
                # normal process
                rows.append(
                    "ATOM".ljust(6, " ")
                    + str(i).rjust(5, " ")
                    + "  "
                    + "O".ljust(3, " ")
                    + " "
                    + "GLY".ljust(3, " ")
                    + " "
                    + chr(65 + chain_idx)
                    + str(j).rjust(4, " ")
                    + "    "
                    + str("%.3f" % (show_shape[0, idx])).rjust(8, " ")
                    + str("%.3f" % (show_shape[1, idx])).rjust(8, " ")
                    + str("%.3f" % (show_shape[2, idx])).rjust(8, " ")
                    + "2.00".rjust(6, " ")
                    + "10.00".rjust(6, " ")
                    + "          "
                    + "O".rjust(2, " ")
                    + "\n"
                )
                Is.append(i)
                i += 1
                j += 1
                k += 1
        if k == 0:
            continue
        i += 1
        rows.append(
            "TER".ljust(6, " ")
            + str(i - 1).rjust(5, " ")
            + "  "
            + "".ljust(3, " ")
            + " "
            + "GLY".ljust(3, " ")
            + " "
            + chr(65 + chain_idx)
            + str(j - 1).rjust(4, " ")
            + "\n"
        )
        for l in Is:
            if l - 1 in Is and l + 1 in Is:
                rows.append(
                    "CONECT".ljust(6, " ")
                    + str(l).rjust(5, " ")
                    + str(l - 1).rjust(5, " ")
                    + str(l + 1).rjust(5, " ")
                    + "\n"
                )
            elif not l - 1 in Is:
                rows.append(
                    "CONECT".ljust(6, " ")
                    + str(l).rjust(5, " ")
                    + str(l + 1).rjust(5, " ")
                    + "\n"
                )
            elif not l + 1 in Is:
                rows.append(
                    "CONECT".ljust(6, " ")
                    + str(l).rjust(5, " ")
                    + str(l - 1).rjust(5, " ")
                    + "\n"
                )
        rows.append("END\n")
        out = open(write_path + "/" + prefix + chain + ".pdb", "wb", 100 * (2**20))
        data = "".join(rows).encode()
        out.write(data)
        out.close()
        chain_idx += 1


def read_gem_as_csv(path, sep="\t"):
    gem = pd.read_csv(
        path,
        sep=sep,
        comment="#",
        dtype=str,
        converters={
            "x": float,
            "y": float,
            "cell_id": str,
            "MIDCount": float,
            "MIDCounts": float,
            "expr": float,
        },
    )
    gem["x"] = gem["x"].astype(int)
    gem["y"] = gem["y"].astype(int)
    # cell column
    if "cell" in gem.columns:
        gem.rename(columns={"cell": "CellID"}, inplace=True)
    elif "cell_id" in gem.columns:
        gem.rename(columns={"cell_id": "CellID"}, inplace=True)
    # count column
    if "MIDCounts" in gem.columns:
        gem.rename(columns={"MIDCounts": "MIDCount"}, inplace=True)
    elif "expr" in gem.columns:
        gem.rename(columns={"expr": "MIDCount"}, inplace=True)
    # gene column
    if "gene" in gem.columns:
        gem.rename(columns={"gene": "geneID"}, inplace=True)

    return gem


def change_last_true(arr):
    # Find the index of the last True value
    last_true_index = np.where(arr == True)[0][-1]

    # Set the value at that index to False
    arr[last_true_index] = False

    return arr


def find_subarray(arr1, arr2):
    n = arr1.shape[0]
    for i in range(n):
        if np.array_equal(arr1[i], arr2):
            return i
    return print("Error! Try to reduce Ngene in MASK function")


def generate_random_rotation_matrices(n):
    # use the Rotation.random method to generate n random rotations
    rotations = R.random(n)
    # convert the rotations to matrices
    matrices = rotations.as_matrix()
    # return the matrices
    return matrices


def get_date_today():
    from datetime import datetime

    # get the current date as a datetime object
    today = datetime.today()
    # format the date as DD-MM-YY
    date_string = today.strftime("%d-%b-%y")
    # print the date string
    return date_string


def load_data(file):
    with open(file, "rb") as f:
        x = pickle.load(f)
    return x


def save_data(data, file):
    with open(file, "wb") as f:
        pickle.dump(data, f)


def generate_id():
    return "".join(random.choices(string.ascii_uppercase + string.digits, k=4))


def legalname(string):
    import re

    # Replace spaces with underscores
    string = re.sub(r"\s+", "_", string)

    # Remove non-alphanumeric characters
    string = re.sub(r"[^a-zA-Z0-9_\-\.]", "", string)

    # Remove leading/trailing underscores
    string = re.sub(r"^_+|_+$", "", string)

    # Remove leading/trailing hyphens
    string = re.sub(r"^-+|-+$", "", string)

    # Remove leading/trailing periods
    string = re.sub(r"^\.+|\.+$", "", string)

    return string


def mean_radius(x):
    center = np.mean(x, axis=0)
    dist = np.linalg.norm(x - center, axis=1)
    return np.mean(dist)  # Return the average distance, i.e. the average radius


def normalizeX(x, method="mean"):
    if method == "L2norm":
        return x / (np.linalg.norm(x) ** 2)
    elif method == "mean":
        r = mean_radius(x)
        return x / r


def craft(
    gem_path,
    species,
    seed,
    threshold_for_gene_filter,
    percent_of_gene_for_rotation_derivation,
    out_path,
    sep,
):
    # set seed
    random.seed(seed)
    np.random.seed(seed)
    #################### INIT ####################
    models = []
    filename, _ = os.path.splitext(gem_path)
    SN = os.path.basename(filename)
    TID = generate_id()
    outpath = out_path + "/" + SN + "_" + TID + "/"

    # make dir
    Path(outpath).mkdir(parents=True, exist_ok=True)
    # start logging
    stdout = sys.stdout
    log_file = open(outpath + "/" + SN + "_" + TID + ".log", "w")
    sys.stdout = log_file

    #################### PROCESSING #####################
    # read input gem
    gem = read_gem_as_csv(gem_path, sep=sep)

    GeneUIDs = list(
        gem.groupby(["geneID"])["MIDCount"]
        .count()
        .reset_index(name="Count")
        .sort_values(["Count"], ascending=False)
        .geneID
    )
    CellUIDs = list(gem.CellID.drop_duplicates())

    print(
        "Speceis: "
        + species
        + "\n"
        + "File Name:"
        + filename
        + "\n"
        + "Seed: "
        + str(seed)
        + "\n"
        + "Cell Number: "
        + str(len(CellUIDs))
        + "\n"
        + "Gene Number: "
        + str(len(GeneUIDs))
        + "\n"
        + "Sample Name: "
        + SN
        + "\n"
        + "Threshold for gene filter is: "
        + str(threshold_for_gene_filter)
        + "\n"
        + "Proportion of genes used for Rotation Derivation is: "
        + str(percent_of_gene_for_rotation_derivation)
        + "\n"
        + "Task ID: "
        + TID
        + "\n"
    )
    Ngene_for_rotation_derivation = int(
        percent_of_gene_for_rotation_derivation * len(GeneUIDs)
    )

    ### processing gtf
    gene_chr = gene_chr(species)

    ##### generate random RM and derive X
    # adata
    data = read_gem(file_path=gem_path, bin_type="cell_bins", sep=sep)
    data.tl.raw_checkpoint()
    adata = stereo_to_anndata(
        data, flavor="scanpy", sample_id=SN, reindex=False, output=None
    )
    # generate and normalize W
    W = genedistribution(gem, CellUIDs, GeneUIDs)
    W, GeneUIDs = filterGenes(W, GeneUIDs, threshold=threshold_for_gene_filter)
    W = normalizeW(W)
    # generate random Rotation Matrix
    RM = generate_random_rotation_matrices(int(W.shape[0] / 2))
    X, W, GeneUIDs = UpdateX(RM, W, GeneUIDs)
    write_pdb(
        X,
        gene_chr,
        GeneUIDs,
        write_path=outpath,
        sp=species,
        seed=seed,
        prefix=SN + "_initial_chr",
    )
    ##### update X
    for loop in range(30):
        # step1: derive Rotation Matrix
        Mask = MASK(
            gem,
            GeneIDs=GeneUIDs,
            CellIDs=CellUIDs,
            Ngene=Ngene_for_rotation_derivation,
        )
        RM, W, CellUIDs, adata = DeriveRotation(W, X, Mask, CellUIDs, adata)
        # step2: update conformation X with RM and W
        try:
            newX, W, GeneUIDs, X = UpdateX(RM, W, GeneUIDs, X)
        except np.linalg.LinAlgError:
            return "numpy.linalg.LinAlgError"
        rmsd, _, _ = numpy_svd_rmsd_rot(
            normalizeX(X, method="mean"), normalizeX(newX, method="mean")
        )
        print(
            "Distance between X_new and X_old for loop "
            + str(loop + 1)
            + " is: "
            + str(rmsd)
        )
        write_pdb(
            newX,
            gene_chr,
            GeneUIDs,
            write_path=outpath,
            sp=species,
            seed=seed,
            prefix=SN + "_updated" + str(loop + 1) + "times_chr",
        )
        ## keep looping until X converges
        X = newX
        if rmsd < 0.25:
            break

    #################### SAVING #####################
    print("Number of total Transcription centers is: " + str(X.shape[0]))

    adata.obsm["Rotation"] = RM
    adata.uns["X"] = pd.DataFrame(X, index=GeneUIDs)
    adata.uns["X"].columns = adata.uns["X"].columns.astype(str)
    adata.uns["W"] = W
    adata.write_h5ad(filename=outpath + "adata.h5ad")

    # stop logging
    sys.stdout = stdout
    log_file.close()


def parse_args():
    parser = argparse.ArgumentParser(description="Cytocraft Main Function")

    parser.add_argument(
        "gem_path",
        type=str,
        help="Path to gem file",
    )
    parser.add_argument(
        "out_path",
        type=str,
        help="Output path to save results",
    )
    parser.add_argument(
        "species",
        type=str,
        help="Species of the input data, e.g. Human, Mouse",
        choices=["Human", "Mouse", "Mice", "Axolotls", "Axolotl", "Monkey"],
    )
    parser.add_argument(
        "-p",
        "--percent",
        type=float,
        help="percent of gene for rotation derivation, default: 0.001",
        default=0.001,
    )
    parser.add_argument(
        "-t",
        "--threshold",
        type=float,
        help="The maximum proportion of np.nans allowed in a column(gene) in W",
        default=0.90,
    )
    parser.add_argument(
        "--sep", type=str, help="input gem file separator", default="\t"
    )
    parser.add_argument(
        "-c",
        "--celltype",
        type=str,
        help="Path of file containing cell types",
    )
    parser.add_argument(
        "--ctkey",
        type=str,
        help="Key of celltype column in the annotation file",
    )
    parser.add_argument(
        "--cikey",
        type=str,
        help="Key of cell id column in the annotation file",
    )
    parser.add_argument(
        "--csep", type=str, help="Annotation file separator", default="\t"
    )
    parser.add_argument(
        "--seed",
        type=int,
        help="Random seed",
    )

    args = parser.parse_args()

    return args


def main():
    #################### SETTINGS ####################
    args = parse_args()
    if args.seed is not None:
        seed = args.seed
    else:
        seed = random.randint(0, 1000)

    #################### RUNNING ####################
    if args.celltype is not None:  # run multiple cell types
        print("Cell types have been set (path is %s)" % args.celltype + "\n")
        gem = read_gem_as_csv(args.gem_path, sep=args.sep)
        celltype = pd.read_csv(args.celltype, sep=args.csep, dtype=str)
        print(celltype[args.ctkey].drop_duplicates())
        # process cell types
        for ct in celltype[args.ctkey].drop_duplicates():
            ct_legal = legalname(ct)
            gem_ct_path = args.gem_path + "." + args.ctkey + "." + ct_legal + ".tsv"
            print(args.cikey)
            if args.cikey is not None:
                cellids = celltype[celltype[args.ctkey] == ct][args.cikey].values
            else:
                cellids = celltype[celltype[args.ctkey] == ct].index.values.astype(str)
                print(cellids)
            gem[gem.CellID.isin(cellids)].to_csv(gem_ct_path, sep="\t", index=False)
            try:
                craft(
                    gem_ct_path,
                    args.species,
                    seed,
                    args.threshold,
                    args.percent,
                    args.out_path,
                    sep="\t",
                )
            except Exception as error:
                print(error)
                print(ct + ": conformation reconstruction failed.")
                continue
            else:
                print(ct + ": conformation reconstruction finished.")
    else:  # run single cell type
        try:
            craft(
                args.gem_path,
                args.species,
                seed,
                args.threshold,
                args.percent,
                args.out_path,
                args.sep,
            )
        except Exception as error:
            print(error)
            print("conformation reconstruction failed.")
        else:
            print("conformation reconstruction finished.")


if __name__ == "__main__":
    main()



import pickle, copy, importlib, math, os
import seaborn as sns
import numpy as np
import pandas as pd
import umap
import scanpy as sc
from tqdm import tqdm

# from pathlib import Path
# from scipy import linalg as LA
# from numpy.linalg import svd, solve, lstsq
# from math import cos, sin
# from scipy.sparse import spmatrix, issparse, csr_matrix
# from anndata import AnnData
# from typing import Optional, Union
# from shapely.geometry import Point, MultiPoint
# from mpl_toolkits.mplot3d import Axes3D
# from scipy.stats import ttest_ind
# import itertools
import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)
from cytocraft.craft import *
from cytocraft import model
from cytocraft import util
from cytocraft.model import BasisShapeModel
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from scipy.cluster.hierarchy import fcluster

# seed = np.random.RandomState(seed=3)


def get_cmap(cmap, ncolor):
    # Define the color map
    cmap = plt.get_cmap(cmap)

    # Get the RGB values of each color in the color map
    hex_values = []
    for i in range(ncolor):
        hex_values.append(mcolors.rgb2hex(cmap(int(256 / ncolor) * i)))

    return hex_values


def plot_RMSD_heatmap(
    D,
    labels,
    cmap="viridis",
    method="average",
    cthresh=4.8,
    font_scale=1,
    cbbox=(0.07, 0.84, 0.04, 0.12),
    lgbox=(0.86, 0.96, 0.01, 0.01),
):
    clustergrid = sns.clustermap(
        D, xticklabels=labels, yticklabels=labels, method=method
    )
    plt.close()

    # Get clusters with threshold
    linkage = clustergrid.dendrogram_row.linkage
    threshold = cthresh
    clusters = fcluster(linkage, threshold, criterion="distance")

    # Create a color map for clusters
    n_clusters = len(np.unique(clusters))
    color_map = sns.color_palette("rainbow", n_clusters)

    # Assign colors to clusters
    cluster_colors = [color_map[c - 1] for c in clusters]

    # Draw plot
    sns.set(rc={"figure.dpi": 300, "figure.figsize": (8, 8)}, font_scale=font_scale)

    ax = sns.clustermap(
        D,
        xticklabels=labels,
        yticklabels=labels,
        cmap=cmap,
        method=method,
        row_colors=[cluster_colors],
        col_colors=[cluster_colors],
        cbar_kws={"ticks": [0, 0.5, 1, 1.5]},
    )
    ax.ax_cbar.set_title("NRMSD")
    ax.ax_cbar.set_position(cbbox)

    # Add legend for row color bar
    from matplotlib.patches import Patch

    lut = dict(zip(clusters, cluster_colors))
    handles = [Patch(facecolor=lut[name]) for name in sorted(lut)]
    legend = plt.legend(
        handles,
        sorted(lut),
        title="Clusters",
        bbox_to_anchor=lgbox,
        bbox_transform=plt.gcf().transFigure,
        loc="upper right",
        ncol=2,
    )
    legend.get_frame().set_facecolor("white")


def plot_network(
    D,
    labels,
    cmap,
    csep="\n",
    solver="barnesHut",
    corder=1,
    N_neighbor=3,
    html="test.html",
    width=1400,
    height=1000,
    edge_scale=150,
    edge_adjust=-60,
):
    from pyvis.network import Network
    from IPython.display import HTML
    import plotly.graph_objs as go
    import itertools
    import networkx as nx
    import plotly.io as pio
    import kaleido
    import json

    # create an empty graph
    G = nx.Graph()

    # add nodes
    N = D.shape[0]
    for i in range(N):
        G.add_node(i, label=labels[i])

    # add edges
    edges = []

    for i in range(N):
        arr = D[i, :]
        min_indices = np.argpartition(arr, int(N_neighbor + 1))[: int(N_neighbor + 1)]
        min_values = arr[min_indices]
        for j in min_indices:
            if not i == j:
                edges.append(
                    (
                        i,
                        int(j),
                        {"weight": float(1 / arr[j] * edge_scale + edge_adjust)},
                    )
                )

    G.add_edges_from(edges)

    # Draw pyvis
    net = Network(width=width, height=height, notebook=True, cdn_resources="remote")
    net.from_nx(G)

    # set node color based on label
    for node in net.nodes:
        node["group"] = node["label"].split(csep)[corder]
        node["color"] = cmap[node["group"]]
        # print(node['color'])

    # show
    net.show_buttons()

    json_obj = {
        "configure": {"enabled": True, "filter": ["nodes", "edges", "physics"]},
        "nodes": {
            "borderWidth": 3,
            "opacity": 1,
            "font": {"size": 15, "strokeWidth": 5},
            "size": 0,
        },
        "edges": {
            "color": {"opacity": 0.7},
            "selfReferenceSize": 0,
            "selfReference": {"size": 0, "angle": 0.7853981633974483},
            "smooth": {"forceDirection": "vertical"},
        },
        "physics": {
            "minVelocity": 0.75,
            "solver": solver,
        },
    }

    net.set_options("const options = " + json.dumps(json_obj))

    net.show(html)


# UMAP
def umap_leiden(
    distance_matrix,
    n_neighbors=15,
    min_dist=0.1,
    resolution=1,
    plot_umap=True,
    seed=99,
    title="",
):
    umap_coord = umap.UMAP(
        metric="precomputed",
        n_neighbors=n_neighbors,
        min_dist=min_dist,
        random_state=seed,
    ).fit_transform(distance_matrix)
    adata = sc.AnnData(pd.DataFrame(np.zeros((len(umap_coord), len(umap_coord)))))
    adata.obsm["umap"] = umap_coord
    sc.pp.neighbors(adata, use_rep="umap")
    sc.tl.leiden(adata, resolution=resolution)
    if plot_umap:
        sc.pl.umap(adata, color=["leiden"], title=title)
    return adata.obs["leiden"]



"""
    Shape basis models.
"""

import numpy as np
import scipy.linalg
from cytocraft import model
from cytocraft.util import norm, normed, vech, from_vech, vc
from cytocraft.model import Scene
from cytocraft import rigid


def to_S_sharp(S):
    F, N = S.shape
    F /= 3
    S_sharp = np.zeros((F, 3 * N))
    S_sharp[:, :N] = S[::3]
    S_sharp[:, N : 2 * N] = S[1::3]
    S_sharp[:, 2 * N : 3 * N] = S[2::3]
    return S_sharp


def to_S(S_sharp):
    F, N = S_sharp.shape
    N /= 3
    S = np.zeros((3 * F, N))
    S[::3] = S_sharp[:, :N]
    S[1::3] = S_sharp[:, N : 2 * N]
    S[2::3] = S_sharp[:, 2 * N : 3 * N]
    return S


def shrink_vec(x, v):
    return np.maximum(x - v, 0.0)


def shrink(X, v):
    U, s, Vt = np.linalg.svd(X, 0)
    s = shrink_vec(s, v)
    return np.dot(U, np.dot(np.diag(s), Vt))


def S_from_Rs_method_1(W, Rs):

    # Create block diagonal matrix R
    R = scipy.linalg.block_diag(*Rs[:, :2, :].tolist())

    # Initialize from method 1.
    S = np.dot(np.linalg.pinv(R), W)

    return S


def S_sharp_from_Rs(W, Rs, init_S_sharp=None):

    if init_S_sharp == None:
        init_S_sharp = to_S_sharp(S_from_Rs_method_1(W, Rs))

    S_sharp = init_S_sharp

    # Use the paramaterization in Dai CVPR 2012 and
    # algorithm from
    # S. Ma, D. Goldfarb, and L. Chen. "Fixed point and bregman iterative
    # methods for matrix rank minimization.", Mathematical Programming,
    # Series A. 2011

    # These values seem to work and be reasonably fast, but were definitely not
    # set in any principled way.
    tau = 1
    mu_bar = 0.0001
    eta = 0.5
    mu = 100000
    xtol = 0.0001 / tau

    # Create block diagonal matrix R
    R = scipy.linalg.block_diag(*Rs[:, :2, :].tolist())

    while mu > mu_bar:
        while True:
            old_P_sharp = S_sharp
            S = to_S(S_sharp)
            g_S_sharp = to_S_sharp(np.dot(R.T, np.dot(R, S) - W))
            Y = S_sharp - tau * g_S_sharp
            S_sharp = shrink(Y, tau * mu)

            # Check for convergence.
            err = np.linalg.norm(old_P_sharp - S_sharp, "fro") / max(
                1.0, np.linalg.norm(S_sharp, "fro")
            )
            if err < xtol:
                break
        # Update mu.
        mu = max(eta * mu, mu_bar)

    return S_sharp


def factor_S_sharp_to_C_and_B(S_sharp, rank):

    # Factor S_sharp into C B_sharp
    C, s, B_sharp = np.linalg.svd(S_sharp)
    s = np.sqrt(s[:rank])
    C = C[:, :rank] * s[np.newaxis, :]
    B_sharp = B_sharp[:rank, :] * s[:, np.newaxis]

    # Reshape into the B matrix.
    B = to_S(B_sharp)

    return C, B


def find_linear_basis_for_vech_Q_k(PI_hat):
    K = PI_hat.shape[1] / 3
    F = PI_hat.shape[0] / 2

    # Find the 2F constraints on vech(Qk)
    A = np.zeros((2 * F, (9 * K * K + 3 * K) / 2))

    for f in range(F):

        # Extract the two rows.
        x_f, y_f = PI_hat[f * 2 : f * 2 + 2]

        # Constraints on vech(Qk) best detailed in equation 2 (rigid case)
        # and the discussion about how this is the same for the nonrigid case
        A[f * 2] = vc(x_f, y_f)
        A[f * 2 + 1] = vc(x_f - y_f, x_f + y_f)

    # Find SVD of the equations.
    U, s, Vh = np.linalg.svd(A)

    # And extract approximate 2*K*K+K dimensional nullspace (See Dai 2021).
    B = Vh[-(2 * K * K - K) :].T

    return B


def find_G_k(M_hat):

    K = M_hat.shape[1] / 3
    F = M_hat.shape[0] / 2

    # Get the linear basis that vech(Q_k) should reside in.
    U = find_linear_basis_for_vech_Q_k(M_hat)

    # 2D subspace of vech style.
    B = np.asarray(
        [
            from_vech(U[:, i], 3 * K, 3 * K, sym=True).flatten()
            for i in range(U.shape[1])
        ]
    ).T

    # Coefficients correspond to taking trace.
    from cvxopt import matrix, solvers

    # Calculate the cost vector.
    non_diag = np.where(np.eye(3 * K).flatten() != 1)[0]
    c = B.copy()
    c[non_diag, :] = 0
    c = c.sum(axis=0)

    # This matrix should be positive semi-definite.
    G = [matrix(-B)]
    h = [matrix(np.zeros((3 * K, 3 * K)))]

    # We fix the first row of the recovered M_hat to have unit length
    # which avoids the trivial solution.
    A = np.dot(vc(M_hat[0], M_hat[0]), U)[np.newaxis, :].copy()
    A = matrix(A)
    b = matrix([[1.0]])

    # Disable progress reporting in solver.
    solvers.options["show_progress"] = False

    # Solve the sdp using cvxopt
    sol = solvers.sdp(matrix(c), Gs=G, hs=h, A=A, b=b)

    # Extract the solution.
    x = np.asarray(sol["x"])

    # Upgrade to vec(Q_k)
    y = np.dot(B, x)

    # And reshape to Q_k
    Q_k = y.reshape((3 * K, 3 * K))

    # Factor into G_k
    Q_k_U, Q_k_s, Q_k_V = np.linalg.svd(Q_k)
    G_k = Q_k_U[:, :3] * np.sqrt(Q_k_s)[np.newaxis, :3]

    # Squared residual objective function as defined in Dai2012.
    def f(g_k):
        G_k = g_k.reshape(3 * K, 3)
        M_k = np.dot(M_hat, G_k)
        M_kx = M_k[0::2]
        M_ky = M_k[1::2]
        term1 = np.square(1 - (M_kx * M_kx).sum(axis=-1) / (M_ky * M_ky).sum(axis=-1))
        term2 = np.square(2 * (M_ky * M_kx).sum(axis=-1) / (M_ky * M_ky).sum(axis=-1))
        err = np.sum(term1 + term2)
        return err

    # Minimize this function with LBFGS
    from scipy.optimize import fmin_l_bfgs_b

    rc = fmin_l_bfgs_b(f, G_k.flatten(), approx_grad=True)
    G_k = rc[0].reshape(3 * K, 3)

    return G_k


def Rs_from_M_k(M_k):
    """Estimates rotations from a column triple of M."""

    n_frames = M_k.shape[0] / 2

    # Estimate scales of each rotation matrix by averaging norms of two rows.
    scales = 0.5 * (norm(M_k[0::2], axis=1) + norm(M_k[1::2], axis=1))

    # Rescale to be close to rotation matrices.
    R = M_k.reshape(n_frames, 2, 3) / scales[:, np.newaxis, np.newaxis]
    Rs = []

    # Upgrade to real rotations.
    for f in range(n_frames):
        Rx = R[f, 0]
        Ry = R[f, 1]
        Rz = normed(np.cross(Rx, Ry))
        U, s, Vh = np.linalg.svd(np.asarray([Rx, Ry, Rz]))
        Rs.append(np.dot(U, Vh))

    return np.asarray(Rs)


def factor(W, n_basis=2, use_method_1=False):
    """Factor's W using

    Dai, Y., Li, H. and He, M. "A Simple Prior-free Method for Non-Rigid
    Structure-from-Motion Factorization". CVPR, 2012

    Note that we use the letter M in place of PI from this paper.
    """

    n_frames = W.shape[0] / 2

    # Estimate 2D translation.
    T = W.mean(axis=-1)
    Ts = np.zeros((n_frames, 3))
    Ts[:, :2] = T.reshape(n_frames, 2)

    # Center observation matrix
    W_cent = W.copy() - T[:, np.newaxis]

    # Factor
    M_hat, B_hat = rigid.factor_matrix(W_cent, n_basis * 3)

    # Recover G_k
    G_k = find_G_k(M_hat)

    # Recover M_k
    M_k = np.dot(M_hat, G_k)

    # Recover R
    Rs = Rs_from_M_k(M_k)

    if use_method_1:

        # Recover S
        S = S_from_Rs_method_1(W_cent, Rs)

        # Build the scene.
        scene = Scene(S=S, Rs=Rs, Ts=Ts)

    else:

        # Recover S_sharp
        S_sharp = S_sharp_from_Rs(W_cent, Rs)

        # Factor to C and B matrices.
        C, B = factor_S_sharp_to_C_and_B(S_sharp, n_basis)

        # Build linear model.
        scene = model.BasisShapeModel(
            Rs, Bs=B.reshape(n_basis, 3, B.shape[1]), C=C, Ts=Ts
        )

    return scene


if __name__ == "__main__":

    # Set the seed.
    np.random.seed(0)

    # Generate some synthetic data.
    n_frames = 200
    gt_model = model.simple_model(n_frames)
    W = gt_model.W

    # Use the Dai algorithm.
    inf_model = factor(W, use_method_1=0)

    # Register to ground truth
    inf_model.register(gt_model)

    model.compare(inf_model, gt_model, visualize=False)



import numpy as np
from cytocraft.util import norm, normed, rotvecs_to_Rs


def point_set_error(Ps, Qs):
    return norm(Ps - Qs, axis=1).mean(axis=-1)


class Scene(object):
    def __init__(self, Ss=None, S=None, Rs=None, Ts=None):

        # Try to use S if Ss is not defined.
        if Ss is None:
            assert S.ndim == 2
            assert S.shape[0] % 3 == 0
            Ss = S.reshape(S.shape[0] / 3, 3, S.shape[1])

        assert Ss.ndim == 3 and Ss.shape[1] == 3

        self.n_frames = Ss.shape[0]
        self.n_points = Ss.shape[2]

        if Rs is None:
            Rs = np.asarray([np.eye(3) for i in range(self.n_frames)])

        if Ts is None:
            Ts = np.zeros((self.n_frames, 3))

        assert Rs.shape[0] == self.n_frames
        assert Rs.shape[1] == Rs.shape[2] == 3

        self.Rs = Rs
        self.Ss = Ss
        self.Ts = Ts

    def copy(self):
        return Scene(Ss=self.Ss, Rs=self.Rs, Ts=self.Ts)

    @property
    def S(self):
        return self.Ss.reshape(self.n_frames * 3, self.n_points)

    @property
    def Ps(self):
        return np.einsum("fij,fjk->fik", self.Rs, self.Ss) + self.Ts[:, :, np.newaxis]

    @property
    def P(self):
        return self.Ps.reshape(self.n_frames * 3, self.n_points)

    @property
    def Ws(self):
        return self.Ps[:, :2]

    @property
    def W(self):
        return self.Ps[:, :2].reshape(self.n_frames * 2, self.n_points)

    @property
    def T(self):
        return self.Ts.reshape(self.n_frames * 3)

    def register(self, gt):
        # Ground truth points.
        Ps_gt = gt.Ps.copy()
        zbar_gt = Ps_gt[:, 2].mean(axis=-1)

        # Find optimal translation
        P1s = self.Ps.copy()
        delta1 = zbar_gt - P1s[:, 2].mean(axis=-1)
        P1s[:, 2] += delta1[:, np.newaxis]
        err1 = point_set_error(P1s, Ps_gt)

        # Reflected optimal translation.
        P2s = self.Ps.copy()
        P2s[:, 2] *= -1
        delta2 = zbar_gt - P2s[:, 2].mean(axis=-1)
        P2s[:, 2] += delta2[:, np.newaxis]
        err2 = point_set_error(P2s, Ps_gt)

        # Figure out which errors are better.
        which = np.argmin(np.asarray([err1, err2]), axis=0)

        # Update z components for when no depth flip is needed.
        idx = which == 0
        self.Ts[idx, 2] += delta1[idx]

        # Update z component for when a depth flip is needed.
        # FIXME: This should be thought about, right now these
        # rotations are becoming reflections, which is probably not
        # optimal.
        idx = which == 1
        self.Ts[idx, 2] *= -1
        self.Ts[idx, 2] += delta2[idx]
        self.Rs[idx, 2] *= -1

    def visualize(self):
        visualize_models([self])

    def __str__(self):
        return "Scene with {n_points} points across {n_frames} frames".format(
            **self.__dict__
        )


class BasisShapeModel(Scene):
    def __init__(self, Rs, Bs, C=None, Ts=None):
        self.n_frames = Rs.shape[0]
        self.n_basis = Bs.shape[0]
        self.n_points = Bs.shape[2]

        if Ts is None:
            Ts = np.zeros((self.n_frames, 3))

        if C is None:
            C = np.ones((self.n_frames, self.n_basis))

        assert (
            Ts.ndim == 2 and Ts.shape[0] == self.n_frames
        ), "Ts matrix the wrong size %s".format((str(Ts.shape)))
        assert C.ndim == 2 and C.shape == (
            self.n_frames,
            self.n_basis,
        ), "C matrix the wrong size"

        self.Rs = Rs
        self.Bs = Bs
        self.Ts = Ts
        self.C = C

    def copy(self):
        return BasisShapeModel(Rs=self.Rs, Bs=self.Bs, C=self.C, Ts=self.Ts)

    @property
    def Ss(self):
        return np.einsum("fk,kij->fij", self.C, self.Bs)

    def __str__(self):
        return "LinearModel with {n_basis} basis shapes of {n_points} points across {n_frames} frames".format(
            **self.__dict__
        )


def generate_rotation_from_optical_axis(D, safety_checks=False):
    """
    D -- a Fx3 vector of directions.

    returns R so that
    Rd = FxN vectors of [0 0 1]
    """

    assert D.ndim == 2 and D.shape[1] == 3

    F = D.shape[0]

    # Space for the rotations.
    R = np.zeros((F, 3, 3))

    # Put the D vectors along the Z axis.
    R[:, 2, :] = normed(D)

    # Make x direction orthogonal in the y=0 plane.
    R[:, 0, 0] = R[:, 2, 2]
    R[:, 0, 2] = -R[:, 2, 0]
    R[:, 0, :] = normed(R[:, 0, :], axis=-1)

    # Make the y direction the cross product.
    # Negation makes this a rotation instead of a reflection.
    R[:, 1, :] = -np.cross(R[:, 0, :], R[:, 2, :], axis=-1)

    if safety_checks:
        for f in range(F):
            if np.linalg.matrix_rank(R[f]) != 3:
                raise Exception("Low rank.")
            if np.linalg.det(R[f]) < 0:
                # Flip one of the rows signs to change this reflection to a rotation matrix.
                R[f, 0, :] *= -1

    return R


def generate_smooth_rotations(n_frames):

    # Create directions that circle in a halo around the origin.
    x = np.cos(np.linspace(0, np.pi * 2, n_frames))
    z = np.ones((n_frames,)) * 1
    y = np.sin(np.linspace(0, np.pi * 2, n_frames))

    # Generate rotations from these directions.
    Rs = generate_rotation_from_optical_axis(np.asarray([x, y, z]).T)

    return Rs


def generate_cube(interval=0.2):

    start = -1 + interval
    stop = 1 - interval

    B = []
    for i_dim1 in range(3):
        for i_dim2 in range(i_dim1 + 1, 3):
            for i_face in range(2):
                for x in np.arange(start, stop, interval):
                    for y in np.arange(start, stop, interval):
                        p = np.ones(3)
                        if i_face:
                            p *= -1
                        p[i_dim1] = x
                        p[i_dim2] = y
                        B += [p]

    B = np.asarray(B).T

    return B


def generate_synthetic_model(n_frames, n_basis, n_points):

    Bs = np.random.randn(n_basis, 3, n_points) * 10
    R = rotvecs_to_Rs(np.random.randn(n_frames, 3))
    C = np.random.randn(n_frames, n_basis)

    return BasisShapeModel(R, Bs, C=C)


def generate_synthetic_rigid_model(n_frames, n_points):

    m = generate_synthetic_model(n_frames, 1, n_points)
    m.C[:] = 1
    return m


def simple_model(n_frames):
    Bs = np.asarray([generate_cube(), generate_cube()])
    Bs[0] = np.random.randn(*Bs[0].shape) * 0.2
    R = generate_smooth_rotations(n_frames)
    C = np.asarray([np.linspace(0, 0.2, n_frames), np.linspace(1, 0.8, n_frames)]).T
    model = BasisShapeModel(R, Bs, C=C)
    return model


def rigid_model(n_frames):
    Bs = np.asarray([generate_cube(interval=0.5)])
    R = generate_smooth_rotations(n_frames)
    model = BasisShapeModel(R, Bs)
    return model


def visualize_models(models):

    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D

    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    plt.ion()

    n_frames = models[0].n_frames

    lines = []
    Pss = [model.Ps for model in models]

    for f in range(n_frames):
        if f == 0:
            for Ps in Pss:
                lines.append(ax.plot(Ps[f, 0], Ps[f, 1], Ps[f, 2], "o")[0])
            ax.set_xlim(-3, 3)
            ax.set_ylim(-3, 3)
            ax.set_zlim(-3, 3)
            ax.set_xlabel("x")
            ax.set_ylabel("y")
            ax.set_zlabel("z")

        else:
            for line, Ps in zip(lines, Pss):
                line.set_data([Ps[f, 0], Ps[f, 1]])
                line.set_3d_properties(Ps[f, 2])

        plt.draw()
        plt.pause(0.1)
        # raw_input()


def visualize_model_shape(model):

    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D

    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    plt.ion()
    lines = []

    lines.append(ax.plot(model.Ss[0, 0], model.Ss[0, 1], model.Ss[0, 2], "o")[0])
    ax.set_xlim(-2, 2)
    ax.set_ylim(-2, 2)
    ax.set_zlim(-2, 2)
    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.set_zlabel("z")

    plt.draw()
    plt.pause(0.1)
    # raw_input()


def compare(inf, gt, visualize=True):
    print("Comparing two models:")
    print("Model 1:", str(inf))
    print("Model 2:", str(gt))

    assert inf.n_frames == gt.n_frames
    assert inf.n_points == inf.n_points

    P_gt = gt.Ps
    P_inf = inf.Ps

    # Get scale of data as defined in Dai2012.
    sigma = np.std(P_gt, axis=-1).mean()

    # This is the average 3D error rescaled by the scale of the data.
    # I think this is close to what is being used in Dai2012.
    err = norm(P_gt - P_inf, axis=1).mean() / sigma
    print("Scaled 3D error", err)

    if visualize:
        visualize_models([gt, inf])

    return err
